/* (c) 2004, Jan D. Wolter, All Rights Reserved. */

/* Various code to support making HTTP request from Backtalk.
 *
 * Can do GET, POST, HEAD queries.  Can do Basic Authentication.
 * Can handle Location redirects.
 *
 * Also handles XML-RPC calls, but it is up to you to provide the xml
 * procedure call document as the query string.
 *
 * Cannot handle compressed content, SSL, or fancier authentication
 * methods.
 */

#include "backtalk.h"
#include "str.h"
#include "vstr.h"
#include "email.h"
#include "http_client.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>


/* FREE_REQUEST - Deallocate a request data structure.
 */

void free_request(struct rq *req)
{
    if (req == NULL) return;
    if (req->host != NULL) free(req->host);
    if (req->path != NULL) free(req->path);
    if (req->query != NULL) free(req->query);
    if (req->user != NULL) free(req->user);
    if (req->password != NULL) free(req->password);
    free(req);
}


/* FREE_RESPONSE - Deallocate a response data structure.
 */

void free_response(struct rp *resp)
{
    if (resp == NULL) return;
    if (resp->reason != NULL) free(resp->reason);
    if (resp->redirect != NULL) free(resp->redirect);
    if (resp->authrealm != NULL) free(resp->authrealm);
    if (resp->type != NULL) free(resp->type);
    if (resp->content != NULL) free(resp->content);
    free(resp);
}


/* FREE_PROXY - Deallocate a proxy data structure.
 */

void free_proxy(struct px *proxy)
{
    if (proxy == NULL) return;
    if (proxy->host != NULL) free(proxy->host);
    free(proxy);
}


/* URI_REQUEST - Given a URL, construct a request data structure and return
 * a pointer to it.  Returns NULL if the URL is not syntactically correct.
 * Method is RM_GET or RM_HEAD, RM_POST, or RM_XML_RPC.
 *
 * URL format is:
 *
 *   http://[<user>:<password>@]<host>[:<port>][<path>][?<query>]
 *
 * The <path> should start with a '/'.  There is no particular support
 * for '#<fragment>' tags, but they might work if you don't mix them with
 * query strings.
 */

struct rq *uri_request(char *uri, int method)
{
    struct rq *req;
    char *b, *p, *q, *x, *c;

    /* URI must start with 'http://' and method must be good */
    if (uri == NULL || strncmp(uri,"http://",7) || method == RM_NONE)
	return NULL;

    /* Make a blank request */
    req= (struct rq *)malloc(sizeof(struct rq));
    req->method= method;
    req->host= NULL;
    req->path= NULL;
    req->query= NULL;
    req->user= NULL;
    req->password= NULL;
    req->redirect= 1;

    /* Pick off the path and query string first */
    b= uri+7;
    p= firstin(b,"/");
    q= firstin(b,"?");

    if (*p == '\0')
	/* Fill in a missing final '/' */
	req->path= strdup("/");
    else
    {
	/* Save path and any query string */
	req->path= strndup(p, q-p);
    }
    if (*q != '\0') req->query= strdup(q+1);

    /* Check if we have a user id */
    c= NULL;
    for (x= b; x < p && *x != '@'; x++)
	if (*x == ':') c= x;
    if (*x == '@')
    {
	if (c == NULL)
	{
	    /* Got an '@' but no ':' before it.  Bad syntax */
	    free_request(req);
	    return NULL;
	}
	req->user= strndup(b, c-b);
	req->password= strndup(c+1, x-c-1);

	b= x+1;
    }

    /* Search for colon between hostname and port number, if we don't already
     * have it. */
    if (c == NULL || *x == '@')
	for (c= b; c < p && *c != ':'; c++)
	    ;
    req->host= strndup(b, c-b);
    req->port= (*c == ':') ? atoi(c+1) : 80;

    return req;
}


/* PROXIFY_REQUEST:  Given a request and the name of a proxy, rewrite the
 * request to route it through the proxy.  Assumes all strings in the req
 * structure are dynamically allocated.
 */

void proxify_request(struct rq *req, struct px *proxy)
{
    char *uri;
    int uri_len;
    char port[16];

    if (proxy == NULL) return;

    uri_len= 7+strlen(req->host)+strlen(req->path);
    if (req->port != 80)
    {
	sprintf(port,":%d",req->port);
	uri_len+= strlen(port);
    }

    /* Construct new path, as 'http://<targhost>:<targport><path>' */
    uri= (char *)malloc(uri_len+1);
    strncpy(uri,"http://",7);
    strcpy(uri+7,req->host);
    if (req->port != 80)
	strcat(uri+7,port);
    strcat(uri+7,req->path);

    /* Free old values */
    free(req->host);
    free(req->path);

    /* Set new values */
    req->host= strdup(proxy->host);
    req->port= proxy->port;
    req->path= uri;
}


/* TCP_OPEN:  Open a TCP connection to the given port on the given host.
 * Returns the descriptor for the new socket if successful, or -1 on failure.
 */

int tcp_open(char *hostname, int portnum)
{
    int sock, a1, a2, a3, a4;
    struct sockaddr_in hostaddr;
    struct protoent *proto;

    if (hostname == NULL) return -1;

    bzero((char *) &hostaddr, sizeof(struct sockaddr_in));

    /* Get protocol */
    if ((proto= getprotobyname("tcp")) == NULL) {
	/* default to standard TCP protocol number */
	proto= malloc(sizeof(struct protoent));
	proto->p_proto= 6;
    }

    /* Create socket */
    if ((sock= socket(AF_INET, SOCK_STREAM, proto->p_proto)) < 0)
	return -1;

    /* Maybe should just call inet_aton() or inet_addr() */
    if (sscanf(hostname, "%d.%d.%d.%d", &a1, &a2, &a3, &a4) == 4)
    {
	/* Connect by IP address */
	uint4 adr;

	/* Construct address */
	adr= htonl((a1 << 24) | (a2 << 16) | (a3 << 8) | a4);
	bcopy((void *) &adr, (void *) &hostaddr.sin_addr, 4);
	hostaddr.sin_family= AF_INET;
	hostaddr.sin_port= htons(portnum);

	/* Try to connect */
	if (connect(sock, (struct sockaddr *) &hostaddr,
		    sizeof(struct sockaddr_in)) < 0) return -1;
    }
    else
    {
	/* Connect by hostname */
	char **addr;
	struct hostent *he;

	/* Lookup list of addresses */
	if ((he= gethostbyname(hostname)) == NULL) return -1;
	hostaddr.sin_family= AF_INET;
	hostaddr.sin_port= htons(portnum);

	/* Try to connect to each */
	for (addr= he->h_addr_list; *addr != NULL; addr++)
	{
	    bcopy((void *) *addr, (void *) &hostaddr.sin_addr, he->h_length);
	    if ((connect(sock, (struct sockaddr *) &hostaddr,
		         sizeof(struct sockaddr_in))) == 0) break;
	}
	if (*addr == NULL) return -1;
    }
    return sock;
}


/* REQUEST_STRING - Given a HTTP request data structure, generate the text
 * to write to the server to make that request, and return it as a string.
 * The returned string is in dynamic memory and should be freed by the caller.
 */

char *request_string(struct rq *req)
{
    char bf[BFSZ];
    int qlen;
    VSTRING out;
    vs_new(&out, 0);

    /* First line */
    switch (req->method)
    {
    case RM_GET: vs_cat(&out,"GET "); break;
    case RM_HEAD: vs_cat(&out,"HEAD "); break;
    case RM_POST: case RM_XML_RPC: vs_cat(&out,"POST "); break;
    }
    vs_cat(&out,req->path);
    if ((req->method == RM_POST || req->method == RM_HEAD) &&
	req->query != NULL)
    {
	*vs_inc(&out)= '?';
	vs_cat(&out,req->query);
    }
    vs_cat(&out," HTTP/1.0\r\n");

    /* User Agent Header */
    sprintf(bf,"User-Agent: Backtalk/%d.%d.%d\r\n",
	VERSION_A,VERSION_B,VERSION_C);
    vs_cat(&out,bf);

    /*vs_cat(&out,"Accept: text/*\r\n");*/

    /* Host header - currently sent only for XML-RPC requests, which
     * seem to require it (while regular http doesn't).
     */
    if (req->method == RM_XML_RPC)
    {
	vs_cat(&out,"Host: ");
	vs_cat(&out,req->host);
	if (req->port != 80)
	{
	    sprintf(bf,":%d",req->port);
	    vs_cat(&out,bf);
	}
	vs_cat(&out,"\r\n");
    }

    /* Authorization header, if an ID was given.
     * We only do Basic Authenication. */
    if (req->user != NULL && req->password != NULL)
    {
	char *enc;

	vs_cat(&out,"Authorization: Basic ");
	sprintf(bf,"%s:%s",req->user,req->password);
	enc= email_b64_encode(bf, strlen(bf), FALSE, FALSE);
	vs_cat(&out,enc);
	free(enc);
	vs_cat_n(&out,"\r\n",2);
    }

    /* Content headers for post request.  We always use urlencoding.
     * No support here for multipart/form-data.  */
    if (req->method == RM_POST)
    {
	vs_cat(&out,"Content-Type: application/x-www-form-urlencoded\r\n");
	qlen= (req->query == NULL) ? 0 : strlen(req->query);
	sprintf(bf,"Content-Length: %d\r\n", qlen);
	vs_cat(&out,bf);
    }
    else if (req->method == RM_XML_RPC)
    {
	vs_cat(&out,"Content-Type: text/xml\r\n");
	qlen= (req->query == NULL) ? 0 : strlen(req->query);
	sprintf(bf,"Content-Length: %d\r\n", qlen);
	vs_cat(&out,bf);
    }

    /* Blank Line marks End of Headers */
    vs_cat_n(&out,"\r\n",2);

    /* Content, if any */
    if ((req->method == RM_POST || req->method == RM_XML_RPC) && qlen > 0)
    {
	vs_cat_n(&out,req->query,qlen);
	vs_cat_n(&out,"\r\n",2);
    }

    /* And return */
    return vs_return(&out);
}


/* READ_RESPONSE - read a response to an HTTP request.  Closes the socket.
 * Returns a response structure.
 */

struct rp *read_response(int sock)
{
    struct rp *resp;
    FILE *fp= fdopen(sock,"r");
    VSTRING hname, hvalue, content;
    int ch, fc;
    int status;
    int badheader= TRUE;
    char *encoding= NULL;

    /* Make us an empty response structure */
    resp= (struct rp *)malloc(sizeof(struct rp));
    resp->status= 0;
    resp->reason= NULL;
    resp->redirect= NULL;
    resp->authrealm= NULL;
    resp->type= NULL;
    resp->content= NULL;

    /* Set up some input buffers */
    vs_new(&hname, 25);
    vs_new(&hvalue, 256);
    vs_new(&content, 1024);

    /* Read in Status Line */
    while ((ch= fgetc(fp)) != EOF && ch != '\n')
	*vs_inc(&hvalue)= ch;
    *vs_inc(&hvalue)= '\0';

    /* Parse the status code and reason out of the status line */
    if (!strncasecmp(hvalue.begin,"HTTP/", 5))
    {
	char *p;
	p= firstin(hvalue.begin, " \t");
	p= firstout(p," \t");
	if (*p >= '0' && *p <= '9')
	{
	    resp->status= atoi(p);
	    p= firstin(p," \t");
	    p= firstout(p," \t");
	    if (*p != '\0')
	    {
		*firstin(p, "\r")= '\0';
		resp->reason= strdup(p);
	    }
	    badheader= FALSE;
	}
    }

    /* If we didn't have a valid status line, treat whole response as
     * content, HTTP 0.9 style. */
    if (badheader)
    {
	/* I guess this counts as success */
	resp->status= 200;
	/* Load header into content array */
	vs_cat(&content, hvalue.begin);
	if (ch != EOF) *vs_inc(&content)= '\n';
	/* Pretend we hit end of headers so header reading loop won't run */
	fc= '\n';
    }
    else
    {
	/* Read first character of first header line */
	fc= fgetc(fp);
    }

    /* Loop through headers */
    while (fc != EOF && fc != '\r' && fc != '\n')
    {
	vs_empty(&hvalue);
	vs_empty(&hname);
	badheader= FALSE;

	/* Read in a header name */
	*vs_inc(&hname)= fc;
	while ((ch= fgetc(fp)) != EOF && ch != ':' && ch != '\n')
	    if (ch != '\r')
		*vs_inc(&hname)= ch;
	if (ch != ':') {fc= fgetc(fp); break;}

	/* Skip white space before value */
	while ((ch= fgetc(fp)) != EOF &&
		(ch == ' ' || ch == '\t' || ch == '\r')) {}
	if (ch == '\n')
	{
	    /* Header with no value, unless there is a continuation */
	    ch= fgetc(fp);
	    if (ch != ' ' && ch != '\t')
	    {
		badheader= TRUE;
		fc= ch;
	    }
	}

	if (!badheader)
	{
	    /* Read in header value */
	    *vs_inc(&hvalue)= ch;
	    while ((ch= fgetc(fp)) != EOF)
	    {
		if (ch == '\r') continue;
		/* A newline followed by non-space/non-tab ends the header */
		if (ch == '\n' && ((ch= fgetc(fp)) != ' ' && ch != '\t'))
		{
		    fc= ch;
		    break;
		}
		*vs_inc(&hvalue)= ch;
	    }
	}

	/* Null terminate things */
	*vs_inc(&hname)= '\0';
	*vs_inc(&hvalue)= '\0';

	/* Have header and value - See if we recognize it */
	if (!strcasecmp(hname.begin,"Location"))
	    resp->redirect= strdup(hvalue.begin);

	else if (!strcasecmp(hname.begin,"WWW-Authenticate"))
	{
	    /* Just ignore anything except basic authentication */
	    if (!strncasecmp(hvalue.begin,"Basic realm=",12))
	    {
		char *e= firstin(hvalue.begin+12,"\"");
		resp->authrealm= strndup(hvalue.begin+12, e-hvalue.begin-12);
	    }
	}

	else if (!strcasecmp(hname.begin,"Content-Encoding"))
	    encoding= strdup(hvalue.begin);

	else if (!strcasecmp(hname.begin,"Content-Type"))
	    resp->type= strdup(hvalue.begin);

	/* Ignore all other response headers */
    }

    if (fc == '\r') fc= fgetc(fp);

    /* Deallocate name/value buffers */
    vs_destroy(&hname);
    vs_destroy(&hvalue);

    /* Read content - we ignore any Content-Length value and just read to
     * EOF, which is basically what the HTTP 1.0 spec says to do */
    if (fc == '\n')
    {
	while ((ch= fgetc(fp)) != EOF)
	    *vs_inc(&content)= ch;
    }

    /* Close that socket */
    fclose(fp);

    /* Save content in response structure, if there was any */
    if (content.p > content.begin) 
	resp->content= vs_return(&content);
    else
	vs_destroy(&content);

    /* Decode content.  Well, maybe someday */
    if (encoding != NULL) free(encoding);

    return resp;
}


/* EXECUTE_REQUEST.  Given non-proxified request structure, and a
 * proxy (which should be NULL if there is none), execute the request
 * and get back a response.  If the response is a location redirect,
 * follow the redirect up to five times.  If unable to connect to the
 * server, or some such other problem, returns NULL.
 *
 * The request will be deallocated by this.
 */

struct rp *execute_request(struct rq *req, struct px *proxy)
{
    char *rstr;
    struct rq *curr= req, *next;
    struct rp *resp;
    int sock;
    int n;

    /* Redirect loop - don't follow more than five redirects */
    for (n=0; n < 5; n++)
    {
	/* Formulate the Request */
	proxify_request(curr, proxy);
	rstr= request_string(curr);

	/* Open the connection */
	sock= tcp_open(curr->host, curr->port);
	if (sock < 0) return NULL;

	/* Send the request */
	write(sock, rstr, strlen(rstr));
	free(rstr);

	/* Get the response (closes socket) */
	resp= read_response(sock);

	/* If not redirecting, get out of this loop */
	if (!req->redirect || (resp->status != 301 && resp->status != 302))
	    break;

	/* Redirecting - Make a new request */
	next= uri_request(resp->redirect, curr->method);
	if (curr->query != NULL)
	{
	    if (next->query != NULL) free(next->query);
	    next->query= curr->query; curr->query= NULL;
	}
	if (curr->user != NULL)
	{
	    if (next->user != NULL) free(next->user);
	    next->user= curr->user; curr->user= NULL;
	}
	if (curr->password != NULL)
	{
	    if (next->password != NULL) free(next->password);
	    next->password= curr->password; curr->password= NULL;
	}

	/* Dispose of old request and response */
	free_request(curr);
	free_response(resp);

	/* And try again... */
	curr= next;
    }

    /* Free our last request */
    free_request(curr);

    return resp;
}


/* PARSE PROXY:  Given a string like "<host>[:<port>]" generate a proxy
 * structure in dynamic memory.  Returns NULL on failure.
 */

struct px *parse_proxy(char *str)
{
    struct px *proxy;
    int port= 80;
    char *c;

    if (str == NULL || *str == '\0') return NULL;

    c= firstin(str,":");
    if (c[0] == ':' && c[1] != '\0')
    {
	if (c[1] < '0' || c[1] > '9') return NULL;
	port= atoi(c+1);
    }

    proxy= (struct px *)malloc(sizeof(struct px));
    proxy->host= strndup(str, c-str);
    proxy->port= port;
}


/* METHOD_CODE:  Given a string value of a method like "POST", return the
 * numeric code.  Returns RM_NONE on bad input.
 */

int method_code(char *str)
{
    if (!strcasecmp(str,"GET")) return RM_GET;
    if (!strcasecmp(str,"POST")) return RM_POST;
    if (!strcasecmp(str,"HEAD")) return RM_HEAD;
    if (!strcasecmp(str,"XML-RPC")) return RM_XML_RPC;
    return RM_NONE;
}
